/*
 * Created on 06.01.2005
 *
 * SVN header information:
 *  $Author: LBST-PF-3\orahn $
 *  $Rev: 2446 $
 *  $Date: 2006-09-12 14:57:25 +0200 (Di, 12 Sep 2006) $
 *  $Id: Data2LayerConnector.java 2446 2006-09-12 12:57:25Z LBST-PF-3\orahn $s
 */
package de.fhOsnabrueck.jump.pirol.utilities.graph;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.util.Assert;
import com.vividsolutions.jump.feature.AttributeType;
import com.vividsolutions.jump.feature.BasicFeature;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureCollection;
import com.vividsolutions.jump.feature.FeatureDataset;
import com.vividsolutions.jump.feature.FeatureSchema;
import com.vividsolutions.jump.workbench.model.Layer;

import de.fhOsnabrueck.jump.pirol.utilities.apiTools.FeatureCollectionTools;
import de.fhOsnabrueck.jump.pirol.utilities.attributes.AttributeInfo;
import de.fhOsnabrueck.jump.pirol.utilities.graph.Delaunay.DelaunayPunkt;
/**
 * 
 * Class to easily exchange JTS geometry types and PIROL geometry
 * types, like punkt or Kante.
 * Offers methods e.g. to convert layers into lists of punkt objects
 * or lists of Kante objects to LinearRing objects.
 * 
 * @author orahn
 *
 * FH Osnabrck - University of Applied Sciences Osnabrck
 * Project PIROL 2005
 * Daten- und Wissensmanagement
 * 
 * @see de.fhOsnabrueck.jump.pirol.utilities.PirolPoint
 * @see de.fhOsnabrueck.jump.pirol.utilities.PirolEdge
 */
public class Data2LayerConnector {
    protected List shapesData = null;
    
    public static final int SHAPETYPE_UNKNOWN = -1;
    public static final int SHAPETYPE_POINT = 0;
    public static final int SHAPETYPE_LINE  = 1;
    public static final int SHAPETYPE_POLYGON = 2;
    public static final int SHAPETYPE_LINEWITHDATA  = 3;
    
    protected int shapeType;
    
    public Data2LayerConnector(List shapesData) {
        this.shapesData = shapesData;
    }
    
    protected void setupShapeType(Object o){
        this.shapeType = Data2LayerConnector.SHAPETYPE_UNKNOWN;
        if (PirolPoint.class.isInstance(o)){
            this.shapeType = Data2LayerConnector.SHAPETYPE_POINT;
            return;
        } else if (LineWithData.class.isInstance(o)) {
            this.shapeType = Data2LayerConnector.SHAPETYPE_LINEWITHDATA;
            return;
        } else if (PirolEdge.class.isInstance(o)) {
            this.shapeType = Data2LayerConnector.SHAPETYPE_LINE;
            return;
        }
    }
    
    protected void setupFeatureSchemaAccordingToShapeType(FeatureSchema fs, AttributeInfo[] additionalAttributes){
        fs.addAttribute("Geometry", AttributeType.GEOMETRY);
        
        if (additionalAttributes!=null){
            for (int i=0; i<additionalAttributes.length; i++){
                fs.addAttribute( additionalAttributes[i].getUniqueAttributeName(), additionalAttributes[i].getAttributeType() );
            }
        }
        
        if (this.shapeType == Data2LayerConnector.SHAPETYPE_POINT){
            fs.addAttribute( "Index", AttributeType.INTEGER );
        } else if (this.shapeType == Data2LayerConnector.SHAPETYPE_LINE){
            fs.addAttribute( "Index P1", AttributeType.INTEGER );
            fs.addAttribute( "Index P2", AttributeType.INTEGER );
            fs.addAttribute( "Index PA", AttributeType.INTEGER );
            fs.addAttribute( "Index PB", AttributeType.INTEGER );
        } else if (this.shapeType == Data2LayerConnector.SHAPETYPE_LINEWITHDATA){}
    }
    
    /**
     * Converts a geometry (with more than 1 Coordinate) to an array of Kante objects.
     *@param poly a Geometry like a LinearRing, a Polygon or a LineString
     *@return array with 0 to n Kante objects
     *@see PirolEdge
     */
    public static PirolEdge[] polygon2KanteList(Geometry poly){
        List<PirolEdge> kanten = new ArrayList<PirolEdge>();
        
        Coordinate[] coords = poly.getCoordinates();
        Coordinate oldCoordinate = coords[0], currentCoordinate;
        
        for (int i=1; i<coords.length; i++){
            currentCoordinate = coords[i];
            kanten.add( new PirolEdge(Data2LayerConnector.coordinate2Punkt(oldCoordinate, PirolPoint.class),Data2LayerConnector.coordinate2Punkt(currentCoordinate, PirolPoint.class)) );
            oldCoordinate = currentCoordinate;
        }
        
        return (PirolEdge[])kanten.toArray(new PirolEdge[0]);
    }
    
    public static Coordinate punkt2Coordinate( PirolPoint p ) throws Exception{
        return PirolPoint.toCoordinate(p);
    }
    
    public static Geometry punkt2Geometry( PirolPoint p, GeometryFactory factory ) throws Exception{
        return factory.createPoint(Data2LayerConnector.punkt2Coordinate(p));
    }
    
    public static LinearRing kantenList2LinearRing( List kanten, GeometryFactory gf ) throws Exception{
        Iterator iter = kanten.iterator();
        Coordinate[] coords = new Coordinate[kanten.size()+1];
        PirolEdge k;
        
        for (int i=0; iter.hasNext(); i++){
            k = (PirolEdge)iter.next();
            coords[i] = Data2LayerConnector.punkt2Coordinate(k.getAnfang());
        }
        coords[kanten.size()] = coords[0];
        // ggf. die erste Koordinate nochmals mit reinnehmen
        LinearRing lr = gf.createLinearRing(coords);
        Assert.isTrue(lr.isRing());
        return lr;
    }
    
    public static Geometry kante2Geometry( PirolEdge k, GeometryFactory factory ) throws Exception{
        Coordinate[] coords = new Coordinate[2];
        coords[0] = Data2LayerConnector.punkt2Coordinate(k.getAnfang());
        coords[1] = Data2LayerConnector.punkt2Coordinate(k.getEnde());
        
        return factory.createLineString(coords);
    }
    
    public static PirolPoint coordinate2Punkt( Coordinate coord, Class punktKlasse ){
        PirolPoint p = null;
        try {
            p = (PirolPoint)punktKlasse.getClassLoader().loadClass(punktKlasse.getName()).newInstance();
            p.setCoordinates(new double[]{coord.x,coord.y});
        } catch (InstantiationException e) {
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }

        return p;
    }
    
    public static PirolPoint[] layer2Punktefeld(Layer layer, Class punktKlasse ) throws InstantiationException, IllegalAccessException, ClassNotFoundException{
        return  Data2LayerConnector.featureList2Punktefeld( FeatureCollectionTools.FeatureCollection2FeatureArray(layer.getFeatureCollectionWrapper().getUltimateWrappee()), punktKlasse );
    }
    
    public static PirolPoint[] featureList2Punktefeld(Feature[] featureList, Class punktKlasse ) throws InstantiationException, IllegalAccessException, ClassNotFoundException{
        ArrayList<PirolPoint> punktefeld = new ArrayList<PirolPoint>();
        Feature[] featureArray = featureList;
        int numFeats = featureArray.length;
        Feature feat;
        PirolPoint p;
        Geometry geom;
        
        for ( int i=0; i<numFeats; i++ ){
            feat = featureArray[i];
            geom = feat.getGeometry();
            
            p = Data2LayerConnector.coordinate2Punkt(geom.getCoordinate(), punktKlasse);
            p.setIndex(feat.getID());
            
            punktefeld.add(p);            
        }
        
        if (punktKlasse.equals(DelaunayPunkt.class)){
            return (DelaunayPunkt[])punktefeld.toArray(new DelaunayPunkt[0]);
        } else if ( punktKlasse.equals(clusterPoint.class)){
            return (clusterPoint[])punktefeld.toArray(new clusterPoint[0]);
        }
        
        return (PirolPoint[])punktefeld.toArray(new PirolPoint[0]);
    }
    /* added by oster */
    /** Converts a Layer into a field of DataPoints.
     *  @param AttributeIndex - The Index of the Attribute which will be stored into the
     * 							DataPoint. 
     * */
    public static List<DataPoint> layer2DataPointField(Layer layer, int AttributeIndex) throws Exception{
        
        ArrayList<DataPoint> punktefeld = new ArrayList<DataPoint>();
        
        FeatureCollection featColl = layer.getFeatureCollectionWrapper().getUltimateWrappee();
        
        Feature[] featureArray = FeatureCollectionTools.FeatureCollection2FeatureArray(featColl);
        Feature feat;
        DataPoint p;
        Geometry geom;
        Object o;
        
        for (int i=0; i<featureArray.length; i++){
            feat = featureArray[i];
            geom = feat.getGeometry();
            p = new DataPoint();
            p.setCoordinates(new double[]{geom.getCoordinate().x,geom.getCoordinate().y});
            p.setIndex(feat.getID());
            // TODO: make sure that attribute-type is double!
            // see FeatureCoolectionTools.isAttributeTypeNumeric()!
            o = feat.getAttribute(AttributeIndex);
            /* adds the attribute as an value: */
            p.setValue(((Double)o).doubleValue());
            punktefeld.add(p);            
        }
        
        return punktefeld;
    }
    /* end added by oster */

    public FeatureCollection putShapesIntoALayer() throws Exception{
        if (this.shapesData == null || this.shapesData.isEmpty()) return null;
        
        FeatureSchema featureSchema = new FeatureSchema();
        
        this.setupShapeType(this.shapesData.get(0));
        if (this.shapeType != SHAPETYPE_LINEWITHDATA){
            this.setupFeatureSchemaAccordingToShapeType(featureSchema, null);
        } else {
            this.setupFeatureSchemaAccordingToShapeType(featureSchema, new AttributeInfo[]{((LineWithData)this.shapesData.get(0)).getAttributeInfo()});
        }
        
        FeatureDataset featureDataset = new FeatureDataset(featureSchema);
        
        Iterator iter = this.shapesData.iterator();
        Feature feat;
        Object obj;
        PirolPoint p;
        PirolEdge k;
        LineWithData l;
        
        GeometryFactory geomFact = new GeometryFactory();
        
        while ( iter.hasNext() ){
            obj = iter.next();
            feat = new BasicFeature(featureSchema);
            
            switch (this.shapeType){
            	case Data2LayerConnector.SHAPETYPE_POINT:
            	    p = (PirolPoint)obj;
            	    feat.setAttribute("Index", new Integer(p.getIndex()));
            	    feat.setGeometry(Data2LayerConnector.punkt2Geometry(p, geomFact));
            	    break;
            	case Data2LayerConnector.SHAPETYPE_LINE:
            	    k = (PirolEdge)obj;
            	    feat.setAttribute("Index P1", new Integer(k.getAnfang().getIndex()));
            	    feat.setAttribute("Index P2", new Integer(k.getEnde().getIndex()));
            	    feat.setAttribute("Index PA", new Integer(k.getPunktIndexA()));
            	    feat.setAttribute("Index PB", new Integer(k.getPunktIndexB()));
            	    feat.setGeometry(Data2LayerConnector.kante2Geometry(k, geomFact));
            	    break;
                case Data2LayerConnector.SHAPETYPE_LINEWITHDATA:
                    l = (LineWithData)obj;
                    feat.setAttribute(l.getAttributeInfo().getUniqueAttributeName(), l.getValue());
                    feat.setGeometry(Data2LayerConnector.kante2Geometry(l, geomFact));
                    break;
            	default:
            	    throw new Exception("Type " + obj.getClass().getName() + " not (yet) supported by "+this.getClass().getName());
            }
    	    featureDataset.add(feat);
        }
        
        return featureDataset;
    }
}
